
#    This file is part of ORDER.
#
#    ORDER is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
# 
# Copyright (C) Maciej Bartkowiak, 2019-2021

# from PyQt5.Qt import Qt as _Qt
# from PyQt5 import QtGui, QtCore, QtWidgets
from PyQt5.QtCore import QThread, QObject, pyqtSignal, pyqtSlot,  QMutex
from PyQt5.QtWidgets import QWidget, QHBoxLayout, QVBoxLayout, QPushButton
from PyQt5.QtWidgets import QSizePolicy, QTableWidget, QTableWidgetItem
import numpy as np
# import copy
from PEAXIS import FlexKernel
from VariablesGUI import TwoWayBox,  VarBox
from ExtendGUI import FlexBar, KnightRider

sample_variables = [
{'Name': 'a',  'Unit':'Angstrom',  'Value':5.840,  'MinValue':0.0,  'MaxValue':1e5,  'Length': 1,  'Type':'float', 'Key':'a'}, 
{'Name': 'b',  'Unit':'Angstrom',  'Value':10.141,  'MinValue':0.0,  'MaxValue':1e5,  'Length': 1,  'Type':'float', 'Key':'b'}, 
{'Name': 'c',  'Unit':'Angstrom',  'Value':6.651,  'MinValue':0.0,  'MaxValue':1e5,  'Length': 1,  'Type':'float', 'Key':'c'}, 
{'Name': 'alpha',  'Unit':'degree',  'Value':90,  'MinValue':0.0,  'MaxValue':180.0,  'Length': 1,  'Type':'float', 'Key':'alpha'}, 
{'Name': 'beta',  'Unit':'degree',  'Value':107.13,  'MinValue':0.0,  'MaxValue':180.0,  'Length': 1,  'Type':'float', 'Key':'beta'}, 
{'Name': 'gamma',  'Unit':'degree',  'Value':90,  'MinValue':0.0,  'MaxValue':180.0,  'Length': 1,  'Type':'float', 'Key':'gamma'}, 
{'Name': 'u',  'Unit':'RLU',  'Value':np.array([0, 1, 0]),
                                      'MinValue':-1e5*np.ones(3),
                                      'MaxValue':1e5*np.ones(3),
                                      'Length': 3,  'Type':'float', 'Key':'u'}, 
{'Name': 'v',  'Unit':'RLU',  'Value':np.array([0, 0,  1]),
                                      'MinValue':-1e5*np.ones(3),
                                      'MaxValue':1e5*np.ones(3),
                                      'Length': 3,  'Type':'float', 'Key':'v'}, 
{'Name': 'misalignment',  'Unit':'degrees',  'Value':np.array([0, 0, 0]),
                                      'MinValue':-360*np.ones(3),
                                      'MaxValue':360*np.ones(3),
                                      'Length': 3,  'Type':'float', 'Key':'misalignment'}, 
]


all_variables = [
{'Name': 'Ei',  'Unit':'eV',  'Value':853,  'MinValue':200.0,  'MaxValue':1200.0,  'Length': 1,  'Type':'float', 'Key':'Ei'}, 
{'Name': 'H',  'Unit':'RLU',  'Value':np.zeros(1),  'MinValue':-1e9*np.ones(1),  'MaxValue':1e9*np.ones(1),
'Length': 1,  'Type':'float', 'Key':'H'}, 
{'Name': 'K',  'Unit':'RLU',  'Value':np.zeros(1),  'MinValue':-1e9*np.ones(1),  'MaxValue':1e9*np.ones(1),
'Length': 1,  'Type':'float', 'Key':'K'}, 
{'Name': 'L',  'Unit':'RLU',  'Value':np.zeros(1),  'MinValue':-1e9*np.ones(1),  'MaxValue':1e9*np.ones(1),
'Length': 1,  'Type':'float', 'Key':'L'}, 
{'Name': 'In-plane vector',  'Unit':'RLU',  'Value':np.zeros(3),  'MinValue':-1e9*np.ones(3),  'MaxValue':1e9*np.ones(3),
'Length': 3,  'Type':'float', 'Key':'lock'}, 
{'Name': 'Q',  'Unit':'AA^(-1)',  'Value':np.zeros(1),  'MinValue':-1e9*np.ones(1),  'MaxValue':1e9*np.ones(1),
'Length': 1,  'Type':'float', 'Key':'Q'}, 
 {'Name': 'Q_par',  'Unit':'AA^(-1)',  'Value':np.zeros(1),  'MinValue':-1e9*np.ones(1),  'MaxValue':1e9*np.ones(1),
'Length': 1,  'Type':'float', 'Key':'Q_par'}, 
{'Name': 'Q_perp',  'Unit':'AA^(-1)',  'Value':np.zeros(1),  'MinValue':-1e9*np.ones(1),  'MaxValue':1e9*np.ones(1),
'Length': 1,  'Type':'float', 'Key':'Q_perp'}, 
{'Name': 'theta',  'Unit':'degree',  'Value':45,  'MinValue':-180,  'MaxValue':180,  'Length': 1,  'Type':'float', 'Key':'theta'}, 
{'Name': 'arm_theta',  'Unit':'degree',  'Value':90,  'MinValue':33.0,  'MaxValue':139.0,  'Length': 1,  'Type':'float', 'Key':'arm_theta'}, 
]

output_variables = [
{'Name': 'zdir',  'Unit':'degree',  'Value':-1.0,  'MinValue':-180.0,  'MaxValue':203.5,  'Length': 1,  'Type':'float', 'Key':'zdir'}, 
{'Name': 'incl',  'Unit':'degree',  'Value':-1.0,  'MinValue':-24.0,  'MaxValue':36.0,  'Length': 1,  'Type':'float', 'Key':'incl'}, 
{'Name': 'rot',  'Unit':'degree',  'Value':-1.0,  'MinValue':-360.0,  'MaxValue':360.0,  'Length': 1,  'Type':'float', 'Key':'rot'}, 
{'Name': 'Cost function',  'Unit':'N/A',  'Value':-1.0,  'MinValue':-1e9,  'MaxValue':1e33,  'Length': 1,  'Type':'float', 'Key':'cost'}, 
]

tabnames = ['CostFunction','ArmTheta','ZDIR','INCL','ROT','H', 'K', 'L', 'Q', 'Q_par', 'Q_perp', 'theta', 'Ei']
class ResultTable(QObject):
    gotvals = pyqtSignal()
    needanupdate = pyqtSignal()
    def __init__(self, master,  nrows =1,  tabnames = []):
        super().__init__(master)
        self.master = master
        self.data = []
        self.useit = []
        self.busy = True
        self.tabnames = tabnames
        ncolumns = len(tabnames)
        self.table = QTableWidget(nrows, ncolumns, self.master)
        self.numcols = ncolumns
        self.table.setSizePolicy(QSizePolicy.MinimumExpanding,QSizePolicy.MinimumExpanding)
        for n in range(self.table.columnCount()):
            self.table.setItem(0, n, QTableWidgetItem(tabnames[n]))
        self.table.cellChanged.connect(self.update_values)
        # self.needanupdate.connect(self.redraw_table)
    def add_row(self, data):
        self.busy = True
        self.table.blockSignals(True)
        lastnit = len(data)
        dummy = self.numcols*['']
        lastrow = self.table.rowCount()
        self.table.insertRow(lastrow)
        for nit, dud in enumerate(self.tabnames):
            try:
                newitem = data[dud]
            except:
                continue
            d = str(round(newitem, 6))
            temp = QTableWidgetItem(str(d).strip("()[]'"))
            self.table.setItem(lastrow, nit, temp)
            dummy[nit] = str(d).strip("()[]'")
        self.data.append(dummy)
        self.busy = False
        self.table.blockSignals(False)
        self.needanupdate.emit()
    @pyqtSlot()
    def clear_table(self):
        self.data = []
        self.useit = []
        for nr in range(1, self.table.rowCount())[::-1]:
            self.table.removeRow(nr)
        self.gotvals.emit()
    def return_values(self):
        final = []
        for nr in range(len(self.useit)):
            if self.useit[nr]:
                if self.data[nr][0] > 0:
                    final.append(self.data[nr])
        return final
    @pyqtSlot(int, int)
    def update_values(self,  row =0,  column=0):
        if self.busy:
            return None
        self.busy = True
        for nr in range(1,  self.table.rowCount()):
            for nc in range(0, self.numcols):
                try:
                    self.data[nr-1][nc] = float(self.table.item(nr, nc).text().strip("()[]'"))
                except:
                    continue
        self.busy = False
        self.needanupdate.emit()
        self.gotvals.emit()
    @pyqtSlot()
    def redraw_table(self):
        if self.busy:
            return None
        self.busy = True
        self.table.blockSignals(True)
        for nr in range(1,  self.table.rowCount()):
            for nc in range(0, self.numcols):
                self.table.item(nr, nc).setText(str(self.data[nr-1][nc]))
        self.busy = False
        self.table.blockSignals(False)
    def assign_fitparams(self, fitparams):
        nums, width,  widtherr,  area,  areaerr = fitparams
        for en,  realnum in enumerate(nums):
            nr = realnum + 1
            for nc in range(5, 9):
                if nc == 5:
                    self.table.item(nr, nc).setText(str(width[en]))
                if nc == 6:
                    self.table.item(nr, nc).setText(str(widtherr[en]))
                if nc == 7:
                    self.table.item(nr, nc).setText(str(area[en]))
                if nc == 8:
                    self.table.item(nr, nc).setText(str(areaerr[en]))


class OrientationTab(QObject):
    values_changed = pyqtSignal()
    pass_input = pyqtSignal(object)
    trigger_processing = pyqtSignal()
    def __init__(self, parent,  forplotting):
        super().__init__(parent)
        self.CK = FlexKernel(None)
        self.figures = []
        # for n in forplotting:
        #     canvas, figure, layout = n
        #     self.figures.append(figure)
        # set up the layout
        self.base = QWidget(parent)
        self.ohlay = QHBoxLayout(self.base)
        self.panel = QWidget(self.base)
        self.plots = QWidget(self.base)
        self.vlay = QHBoxLayout(self.panel)
        self.pvlay = QVBoxLayout(self.plots)
        self.tophalf = QWidget(self.base)
        self.hlay1 = QVBoxLayout(self.tophalf)
        self.bottomhalf = QWidget(self.base)
        self.hlay2 = QVBoxLayout(self.bottomhalf)
        self.ohlay.addWidget(self.panel)
        self.ohlay.addWidget(self.plots)
        self.vlay.addWidget(self.tophalf)
        self.vlay.addWidget(self.bottomhalf)
        # new parts
        self.progbar = FlexBar(self.base)
        self.oscillator = KnightRider(self.progbar, parent = self.base)
        # end of new parts
        # for n in forplotting:
        #     canvas, figure, layout = n
        #     self.pvlay.addWidget(canvas)
        # now add the components
        self.box1 = VarBox(self.base, sample_variables,  "Sample (read-only)", prec_override = 5)
        self.hlay2.addWidget(self.box1.base)
        self.box2 = TwoWayBox(self.base, all_variables,  "All parameters", prec_override = 5)
        self.hlay1.addWidget(self.box2.base)
        self.box3 = VarBox(self.base, output_variables,  "Output (read-only)", prec_override = 5)
        self.hlay2.addWidget(self.box3.base)
        # self.box3 = VarBox(self.base, rixs_variables,  "Scattering", prec_override = 5)
        trigger = QPushButton("Calculate!",  self.bottomhalf)
        trigger.clicked.connect(self.calculate)
        self.trigger = trigger
        self.hlay2.addWidget(trigger)
        self.hlay2.addWidget(self.progbar)
        # self.hlay2.addWidget(canvas)
        self.resultviewer = ResultTable(self.base, tabnames = tabnames)
        self.hlay1.addWidget(self.resultviewer.table)
        self.sample_vals = self.box1.returnValues()
        self.all_vals = self.box2.returnValues()
        self.all_flags = self.box2.returnFlags()
        # self.rixs_vals = self.box3.returnValues()
        # and link the signals to the slots to automatise the operation
        self.box1.values_changed.connect(self.on_parchange)
        self.box2.values_changed.connect(self.on_parchange)
        self.box1.disableChanges()
        self.box3.disableChanges()
        # self.box3.values_changed.connect(self.on_parchange)
        self.CK.motors_result.connect(self.output)
        self.CK.other_results.connect(self.output_others)
        self.CK.table_results.connect(self.output_table)
        #self.CK.hkl_result.connect(self.hkl_output)
        #self.CK.hkl_limits.connect(self.added_hkl_output)        
        #
        self.pass_input.connect(self.CK.parse_input)
        self.trigger_processing.connect(self.CK.find_angles)
        self.corethread = QThread()
        # if app is not None:
        #     app.aboutToQuit.connect(self.corethread.quit)
        #     app.aboutToQuit.connect(self.cleanup)
        self.base.destroyed.connect(self.corethread.quit)
        # self.core.read_tdb_profile()
        self.CK.moveToThread(self.corethread)
        self.corethread.start()
    @pyqtSlot(object)
    def updateSample(self, input):
        vlist,  vdict = input[0],  input[1]
        self.box1.enableChanges()
        self.box1.takeValues(vlist, vdict)
        self.box1.disableChanges()
    @pyqtSlot()
    def calculate(self):
        self.oscillator.start()
        self.trigger.setEnabled(False)
        self.resultviewer.clear_table()
        self.pass_input.emit([self.sample_vals,  self.all_vals,  self.all_flags])
        # self.CK.parse_input([self.sample_vals,  self.all_vals,  self.all_flags])
        self.trigger_processing.emit()
        # self.CK.find_angles()
    def on_parchange(self):
        self.sample_vals = self.box1.returnValues()
        self.all_vals = self.box2.returnValues()
        self.all_flags = self.box2.returnFlags()
        return None
        # self.rixs_vals = self.box3.returnValues()
        # self.calculate()
    @pyqtSlot(object)
    def output(self, results):
        # self.sample_vals = self.box1.returnValues()
        # self.instrument_vals = self.box2.returnValues()
        # self.rixs_vals = self.box3.returnValues()
        self.trigger.setEnabled(True)
        self.oscillator.stop()
        zdir, incl, rot = results[0]
        cfun = results[1]
        vnames = ['zdir', 'incl', 'rot', 'cost']
        vvals = {
                    'zdir' : [zdir], 
                    'incl':[incl], 
                    'rot': [rot], 
                    'cost': [cfun]
        }
        # print(results)
        self.box3.enableChanges()
        self.box3.takeValues(vnames,  vvals)
        self.box3.disableChanges()
        # now some added plotting
        # VD = self.CK.plot_vectors()
        # plot_vectors_projected(VD,  projection = (1, 0),  fig = self.figures[0])
        # plot_vectors_projected(VD,  projection = (1, 2),  fig = self.figures[1])
        # plot_vectors_projected(VD,  projection = (0, 2),  fig = self.figures[2])    @pyqtSlot(object)
    @pyqtSlot(object)
    def output_table(self, results):
        for vdict in results:
            self.resultviewer.add_row(vdict)
    @pyqtSlot(object)
    def output_others(self, results):
        vnames = []
        vvals = {}
        for kk in results.keys():
            strk = str(kk)
            vnames.append(strk)
            try:
                vlen = len(results[kk])
            except:
                vvals[strk] = [results[kk]]
            else:
                vvals[strk] = results[kk]
        # self.box2.enableChanges()
        self.box2.takeValues(vnames,  vvals)
        # self.box2.disableChanges()
    def hkl_output(self, results):
        # self.sample_vals = self.box1.returnValues()
        # self.instrument_vals = self.box2.returnValues()
        # self.rixs_vals = self.box3.returnValues()
        h,  k, l, u, v, w, qm, qp= results
        vnames = ['HKL1',  'HKL2', 'Q', 'Q_perp']
        vvals = {'HKL1' : [h,  k,  l], 
                     'HKL2' : [u, v, w], 
                     'Q' : [qm], 
                     'Q_perp':[qp]
        }
        # print(results)
        self.box3.enableChanges()
        self.box3.takeValues(vnames,  vvals)
        self.box3.disableChanges()
    def added_hkl_output(self, results):
        # self.sample_vals = self.box1.returnValues()
        # self.instrument_vals = self.box2.returnValues()
        # self.rixs_vals = self.box3.returnValues()
        hkl_min,  hkl_max = results
        newvals = 0.5*np.abs(hkl_max-hkl_min)
        vnames = ['HKL_var1',  'HKL_var2']
        vvals = {'HKL_var1' : newvals[0:3], 
                    'HKL_var2' : newvals[3:6], 
        }
        # print(results)
        self.box3.enableChanges()
        self.box3.takeValues(vnames,  vvals)
        self.box3.disableChanges()

def plot_vectors_projected(vdict,  projection = (1, 0), outFile = "", fig = None):
    if fig == None:
        fig = mpl.figure(figsize = [7.0, 6.0], dpi=75, frameon = False)
        trigger = True
    else:
        fig.clear()
        trigger = False
    ax = fig.add_subplot(111)
    ax.grid(True)
    ax.set_xlim([1.0, -1.0])
    ax.set_ylim([-1.0, 1.0])
    axlabels = ['X', 'Y', 'Z']
    title = axlabels[projection[1]] + ' vs. ' + axlabels[projection[0]] 
    # ax.tick_params(bottom = 'off', top = 'off', left = 'off', right = 'off',
    #                labelbottom = 'off', labeltop = 'off', labelleft = 'off', labelright = 'off')
    ax.set_title(title +" scattering geometry")
    ax.set_ylabel(axlabels[projection[1]])
    # ax.text(0.12, 0.70, "This is not the same coverage calculator as before.")
    names = vdict.keys()
    for nv in names:
        start,  end = vdict[nv]
        spoint = (start[projection[0]],  start[projection[1]])
        epoint = (end[projection[0]],  end[projection[1]])
        ax.plot([spoint[0], epoint[0]], [spoint[1],  epoint[1]],  label = nv)
        # ax.annotate(nv, epoint, xytext = spoint,
        #         arrowprops={'arrowstyle':'-|>', 'mutation_scale':10, 'linewidth':2.0})
                #, 'shrinkA':5, 'shrinkB':7, 'mutation_scale':20, 'mutation_aspect':1.3})
                #, 'width':1.5, 'frac':0.2, 'headwidth':5})
    ax.legend(loc=0)
    if not outFile:
        if trigger:
            mpl.show()
        else:
            fig.canvas.draw()
    else:
        mpl.savefig(outFile)
        mpl.close()
