
#    This file is part of ORDER.
#
#    ORDER is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
# 
# Copyright (C) Maciej Bartkowiak, 2019-2021

# from PyQt5.Qt import Qt as _Qt
# from PyQt5 import QtGui, QtCore, QtWidgets
from PyQt5.QtCore import QThread, QObject, pyqtSignal, pyqtSlot,  QMutex
from PyQt5.QtWidgets import QWidget, QHBoxLayout, QVBoxLayout
import numpy as np
# import copy
from PEAXIS import ReverseKernel
from VariablesGUI import VarBox

sample_variables = [
{'Name': 'a',  'Unit':'Angstrom',  'Value':5.840,  'MinValue':0.0,  'MaxValue':1e5,  'Length': 1,  'Type':'float', 'Key':'a'}, 
{'Name': 'b',  'Unit':'Angstrom',  'Value':10.141,  'MinValue':0.0,  'MaxValue':1e5,  'Length': 1,  'Type':'float', 'Key':'b'}, 
{'Name': 'c',  'Unit':'Angstrom',  'Value':6.651,  'MinValue':0.0,  'MaxValue':1e5,  'Length': 1,  'Type':'float', 'Key':'c'}, 
{'Name': 'alpha',  'Unit':'degree',  'Value':90,  'MinValue':0.0,  'MaxValue':180.0,  'Length': 1,  'Type':'float', 'Key':'alpha'}, 
{'Name': 'beta',  'Unit':'degree',  'Value':107.13,  'MinValue':0.0,  'MaxValue':180.0,  'Length': 1,  'Type':'float', 'Key':'beta'}, 
{'Name': 'gamma',  'Unit':'degree',  'Value':90,  'MinValue':0.0,  'MaxValue':180.0,  'Length': 1,  'Type':'float', 'Key':'gamma'}, 
{'Name': 'u',  'Unit':'RLU',  'Value':np.array([0, 1, 0]),
                                      'MinValue':-1e5*np.ones(3),
                                      'MaxValue':1e5*np.ones(3),
                                      'Length': 3,  'Type':'float', 'Key':'u'}, 
{'Name': 'v',  'Unit':'RLU',  'Value':np.array([0, 0,  1]),
                                      'MinValue':-1e5*np.ones(3),
                                      'MaxValue':1e5*np.ones(3),
                                      'Length': 3,  'Type':'float', 'Key':'v'}, 
{'Name': 'misalignment',  'Unit':'degrees',  'Value':np.array([0, 0, 0]),
                                      'MinValue':-360*np.ones(3),
                                      'MaxValue':360*np.ones(3),
                                      'Length': 3,  'Type':'float', 'Key':'misalignment'}, 
]


input_variables = [
{'Name': 'HKL',  'Unit':'RLU',  'Value':np.zeros(3),  'MinValue':-1e9*np.ones(3),  'MaxValue':1e9*np.ones(3),
'Length': 3,  'Type':'float', 'Key':'HKL'}, 
# {'Name': 'Q',  'Unit':'AA^(-1)',  'Value':np.zeros(1),  'MinValue':-1e9*np.ones(1),  'MaxValue':1e9*np.ones(1),
# 'Length': 1,  'Type':'float', 'Key':'Q'}, 
# {'Name': 'Q_perp',  'Unit':'AA^(-1)',  'Value':np.zeros(1),  'MinValue':-1e9*np.ones(1),  'MaxValue':1e9*np.ones(1),
# 'Length': 1,  'Type':'float', 'Key':'Q_perp'}, 
# {'Name': 'Q_par',  'Unit':'AA^(-1)',  'Value':np.zeros(1),  'MinValue':-1e9*np.ones(1),  'MaxValue':1e9*np.ones(1),
# 'Length': 1,  'Type':'float', 'Key':'Q_par'}, 
{'Name': 'Ei',  'Unit':'eV',  'Value':853,  'MinValue':200.0,  'MaxValue':1200.0,  'Length': 1,  'Type':'float', 'Key':'Ei'}, 
{'Name': 'dE',  'Unit':'eV',  'Value':0.0,  'MinValue':-100.0,  'MaxValue':100.0,  'Length': 1,  'Type':'float', 'Key':'dE'}, 
{'Name': 'In-plane vector (optional)',  'Unit':'RLU',  'Value':np.zeros(3),  'MinValue':-1e9*np.ones(3),  'MaxValue':1e9*np.ones(3),
'Length': 3,  'Type':'float', 'Key':'lock'}, 
]

output_variables = [
{'Name': 'arm_theta',  'Unit':'degree',  'Value':90,  'MinValue':33.0,  'MaxValue':139.0,  'Length': 1,  'Type':'float', 'Key':'arm_theta'}, 
{'Name': 'xpos',  'Unit':'mm',  'Value':5.0,  'MinValue':-50.0,  'MaxValue':50.0,  'Length': 1,  'Type':'float', 'Key':'xpos'}, 
{'Name': 'ypos',  'Unit':'mm',  'Value':5.0,  'MinValue':-50.0,  'MaxValue':50.0,  'Length': 1,  'Type':'float', 'Key':'ypos'}, 
{'Name': 'zpos',  'Unit':'mm',  'Value':5.0,  'MinValue':-250.0,  'MaxValue':50.0,  'Length': 1,  'Type':'float', 'Key':'xpos'}, 
{'Name': 'zdir',  'Unit':'degree',  'Value':45.0,  'MinValue':-180.0,  'MaxValue':203.5,  'Length': 1,  'Type':'float', 'Key':'zdir'}, 
{'Name': 'incl',  'Unit':'degree',  'Value':0.0,  'MinValue':-24.0,  'MaxValue':36.0,  'Length': 1,  'Type':'float', 'Key':'incl'}, 
{'Name': 'rot',  'Unit':'degree',  'Value':0.0,  'MinValue':-360.0,  'MaxValue':360.0,  'Length': 1,  'Type':'float', 'Key':'rot'}, 
]

output_hkl = [
{'Name': 'HKL',  'Unit':'RLU',  'Value':np.zeros(3),  'MinValue':-1e9*np.ones(3),  'MaxValue':1e9*np.ones(3),
'Length': 3,  'Type':'float', 'Key':'HKL'}, 
{'Name': 'dHKL',  'Unit':'RLU',  'Value':np.zeros(3),  'MinValue':-1e9*np.ones(3),  'MaxValue':1e9*np.ones(3),
'Length': 3,  'Type':'float', 'Key':'HKL_var'}, 
]

class ReverseSampleTab(QObject):
    values_changed = pyqtSignal()
    def __init__(self, parent,  forplotting):
        super().__init__(parent)
        self.CK = ReverseKernel(parent)
        self.figures = []
        # for n in forplotting:
        #     canvas, figure, layout = n
        #     self.figures.append(figure)
        # set up the layout
        self.base = QWidget(parent)
        self.ohlay = QHBoxLayout(self.base)
        self.panel = QWidget(self.base)
        self.plots = QWidget(self.base)
        self.vlay = QVBoxLayout(self.panel)
        self.pvlay = QVBoxLayout(self.plots)
        self.tophalf = QWidget(self.base)
        self.hlay1 = QHBoxLayout(self.tophalf)
        self.bottomhalf = QWidget(self.base)
        self.hlay2 = QHBoxLayout(self.bottomhalf)
        self.ohlay.addWidget(self.panel)
        self.ohlay.addWidget(self.plots)
        self.vlay.addWidget(self.tophalf)
        self.vlay.addWidget(self.bottomhalf)
        # for n in forplotting:
        #     canvas, figure, layout = n
        #     self.pvlay.addWidget(canvas)
        # now add the components
        self.box1 = VarBox(self.base, sample_variables,  "Sample (read-only)", prec_override = 5)
        self.hlay1.addWidget(self.box1.base)
        self.box2 = VarBox(self.base, input_variables,  "Destination (Input)", prec_override = 5)
        self.hlay1.addWidget(self.box2.base)
        # self.box3 = VarBox(self.base, rixs_variables,  "Scattering", prec_override = 5)
        # self.hlay2.addWidget(self.box3.base)
        self.box4 = VarBox(self.base, output_variables,  "Motor positions (Output)", prec_override = 5)
        self.hlay2.addWidget(self.box4.base)
        self.box3 = VarBox(self.base, output_hkl,  "HKL from new motors (Output)", prec_override = 5)
        self.hlay2.addWidget(self.box3.base)
        # self.hlay2.addWidget(canvas)
        self.sample_vals = self.box1.returnValues()
        self.input_vals = self.box2.returnValues()
        # self.rixs_vals = self.box3.returnValues()
        # and link the signals to the slots to automatise the operation
        self.box1.values_changed.connect(self.on_parchange)
        self.box2.values_changed.connect(self.on_parchange)
        self.box1.disableChanges()
        self.box4.disableChanges()
        self.box3.disableChanges()
        self.box4.values_changed.connect(self.postcalculate)
        # self.box3.values_changed.connect(self.on_parchange)
        self.CK.motors_result.connect(self.output)
        self.CK.hkl_result.connect(self.hkl_output)
        self.CK.hkl_limits.connect(self.added_hkl_output)
    @pyqtSlot(object)
    def updateSample(self, input):
        vlist,  vdict = input[0],  input[1]
        self.box1.enableChanges()
        self.box1.takeValues(vlist, vdict)
        self.box1.disableChanges()
    @pyqtSlot()
    def postcalculate(self):
        self.CK.verify_hkl(self.box4.returnValues()[1])
    def calculate(self):
        self.CK.parse_input([self.sample_vals,  self.input_vals])
        self.CK.calc_Q()
        self.CK.calc_Rmatrix()
        self.CK.find_angles()
    def on_parchange(self):
        self.sample_vals = self.box1.returnValues()
        self.input_vals = self.box2.returnValues()
        # self.rixs_vals = self.box3.returnValues()
        self.calculate()
    @pyqtSlot(object)
    def output(self, results):
        # self.sample_vals = self.box1.returnValues()
        # self.instrument_vals = self.box2.returnValues()
        # self.rixs_vals = self.box3.returnValues()
        atheta, zdir, incl, rot = results
        vnames = ['arm_theta', 'zdir', 'incl', 'rot']
        vvals = {'arm_theta' : [atheta], 
                    'zdir' : [zdir], 
                    'incl':[incl], 
                    'rot': [rot]
        }
        # print(results)
        self.box4.enableChanges()
        self.box4.takeValues(vnames,  vvals)
        self.box4.disableChanges()
        # now some added plotting
        # VD = self.CK.plot_vectors()
        # plot_vectors_projected(VD,  projection = (1, 0),  fig = self.figures[0])
        # plot_vectors_projected(VD,  projection = (1, 2),  fig = self.figures[1])
        # plot_vectors_projected(VD,  projection = (0, 2),  fig = self.figures[2])    @pyqtSlot(object)
    def hkl_output(self, results):
        # self.sample_vals = self.box1.returnValues()
        # self.instrument_vals = self.box2.returnValues()
        # self.rixs_vals = self.box3.returnValues()
        h,  k, l = results
        vnames = ['HKL']
        vvals = {'HKL' : [h,  k,  l]
        }
        # print(results)
        self.box3.enableChanges()
        self.box3.takeValues(vnames,  vvals)
        self.box3.disableChanges()
    def added_hkl_output(self, results):
        # self.sample_vals = self.box1.returnValues()
        # self.instrument_vals = self.box2.returnValues()
        # self.rixs_vals = self.box3.returnValues()
        hkl_min,  hkl_max = results
        newvals = 0.5*np.abs(hkl_max-hkl_min)
        vnames = ['HKL_var']
        vvals = {'HKL_var' : newvals
        }
        # print(results)
        self.box3.enableChanges()
        self.box3.takeValues(vnames,  vvals)
        self.box3.disableChanges()

def plot_vectors_projected(vdict,  projection = (1, 0), outFile = "", fig = None):
    if fig == None:
        fig = mpl.figure(figsize = [7.0, 6.0], dpi=75, frameon = False)
        trigger = True
    else:
        fig.clear()
        trigger = False
    ax = fig.add_subplot(111)
    ax.grid(True)
    ax.set_xlim([1.0, -1.0])
    ax.set_ylim([-1.0, 1.0])
    axlabels = ['X', 'Y', 'Z']
    title = axlabels[projection[1]] + ' vs. ' + axlabels[projection[0]] 
    # ax.tick_params(bottom = 'off', top = 'off', left = 'off', right = 'off',
    #                labelbottom = 'off', labeltop = 'off', labelleft = 'off', labelright = 'off')
    ax.set_title(title +" scattering geometry")
    ax.set_ylabel(axlabels[projection[1]])
    # ax.text(0.12, 0.70, "This is not the same coverage calculator as before.")
    names = vdict.keys()
    for nv in names:
        start,  end = vdict[nv]
        spoint = (start[projection[0]],  start[projection[1]])
        epoint = (end[projection[0]],  end[projection[1]])
        ax.plot([spoint[0], epoint[0]], [spoint[1],  epoint[1]],  label = nv)
        # ax.annotate(nv, epoint, xytext = spoint,
        #         arrowprops={'arrowstyle':'-|>', 'mutation_scale':10, 'linewidth':2.0})
                #, 'shrinkA':5, 'shrinkB':7, 'mutation_scale':20, 'mutation_aspect':1.3})
                #, 'width':1.5, 'frac':0.2, 'headwidth':5})
    ax.legend(loc=0)
    if not outFile:
        if trigger:
            mpl.show()
        else:
            fig.canvas.draw()
    else:
        mpl.savefig(outFile)
        mpl.close()
